from unittest.mock import patch
from uuid import uuid4

from django.contrib.gis.geos import LineString
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element, ElementPath
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestBulkElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_run = cls.worker_version.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.custom_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.local_worker_run = WorkerRun.objects.get(version=cls.custom_version, process__creator=cls.user)
        cls.payload = {
            "worker_run_id": str(cls.local_worker_run.id),
            "elements": [
                {
                    "name": "A",
                    "type": "act",
                    "polygon": [
                        [0, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    "name": "B",
                    "type": "surface",
                    # Reuses the same zone
                    "polygon": [
                        [0, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    "name": "C",
                    "type": "surface",
                    "polygon": [
                        [0, 0],
                        [0, 9],
                        [9, 9],
                        [9, 0],
                        [0, 0],
                    ]
                }
            ]
        }

    def setUp(self):
        super().setUp()
        self.element = self.corpus.elements.get(name="Volume 2, page 1v")
        self.rotated_page = Element.objects.create(corpus=self.corpus, type=self.element.type, image=self.element.image, polygon=self.element.polygon, name="rotated_page", rotation_angle=90)
        self.mirrored_page = Element.objects.create(corpus=self.corpus, type=self.element.type, image=self.element.image, polygon=self.element.polygon, name="mirrored_page", mirrored=True)

    def test_bulk_create_requires_login(self):
        response = self.client.post(
            reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
            data=self.payload,
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_bulk_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
            data=self.payload,
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_bulk_create_writable_corpus(self, filter_rights_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
            data=self.payload,
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_bulk_create_requires_parent_zone(self):
        self.element.image = None
        self.element.polygon = None
        self.element.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
            data={"elements": []},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "element": ["Cannot create child elements on an element without a zone."]
        })

    def test_bulk_create_requires_elements(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data={
                    "worker_run_id": str(self.local_worker_run.id),
                    "elements": [],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"elements": {"non_field_errors": ["This list may not be empty."]}})

    def test_bulk_create_missing_types(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data={
                    "worker_run_id": str(self.local_worker_run.id),
                    "elements": [
                        {
                            "name": "nope",
                            "type": "nope",
                            "polygon": [[0, 0], [0, 42], [13, 37]],
                        }
                    ]
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "non_field_errors": ["Element types with slugs nope do not exist in the parent element's corpus"]
        })

    def test_bulk_create_multiple_parent_paths(self):
        parent = self.corpus.elements.create(
            name="Parent 2",
            type=self.corpus.types.first(),
        )
        self.element.add_parent(parent)
        element_path1, element_path2 = self.element.paths.order_by("path__0")

        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data=self.payload,
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        a, b, c = Element \
            .objects \
            .get_descending(self.element.id) \
            .filter(type__slug__in=("act", "surface")) \
            .order_by("name")

        a_path1, a_path2 = a.paths.order_by("path__0")
        b_path1, b_path2 = b.paths.order_by("path__0")
        c_path1, c_path2 = c.paths.order_by("path__0")

        self.assertListEqual(a_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(a_path2.path, element_path2.path + [self.element.id])
        self.assertListEqual(b_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(b_path2.path, element_path2.path + [self.element.id])
        self.assertListEqual(c_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(c_path2.path, element_path2.path + [self.element.id])

        self.assertEqual(a_path1.ordering, 0)
        self.assertEqual(a_path2.ordering, 0)
        self.assertEqual(b_path1.ordering, 1)
        self.assertEqual(b_path2.ordering, 1)
        self.assertEqual(c_path1.ordering, 2)
        self.assertEqual(c_path2.ordering, 2)

    def test_bulk_create_top_level(self):

        # Create a top level page, without any path
        element = self.corpus.elements.create(
            name="Top level page",
            type=self.corpus.types.get(slug="page"),
            image=self.element.image,
            polygon=self.element.polygon,
        )
        self.assertFalse(element.paths.exists())

        # Try to create a sub element on that zone
        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(element.id)}),
                data={
                    "worker_run_id": str(self.local_worker_run.id),
                    "elements": [
                        {
                            "name": f"surface {i}",
                            "type": "surface",
                            "polygon": [
                                [i * 100, i * 100],
                                [(i + 1) * 100, i * 100],
                                [(i + 1) * 100, (i + 1) * 100],
                                [i * 100, (i + 1) * 100],
                            ]
                        }
                        for i in range(1, 5)
                    ]
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # Check the created elements are children of element
        data = response.json()
        self.assertEqual(len(data), 4)
        paths = ElementPath.objects.filter(element_id__in=[elt["id"] for elt in data], path=[element.id])
        self.assertEqual(paths.count(), 4)

    def test_bulk_create_negative_polygon(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                {
                    "worker_run_id": str(self.local_worker_run.id),
                    "elements": [
                        {
                            "name": "Blah",
                            "type": "surface",
                            "polygon": [[0, 10], [10, 20], [30, 40], [50, 60], [0, -10]]
                        }
                    ]
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "elements": [
                {
                    "polygon": {
                        "4": {
                            "1": ["Ensure this value is greater than or equal to 0."]
                        }
                    }
                }
            ]
        })

    def test_bulk_create_inherit_rotation(self):
        """
        Created children elements inherit their parent's rotation angle
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-bulk-create", kwargs={"pk": str(self.rotated_page.id)}),
            data=self.payload,
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_elements = Element.objects.get_descending(self.rotated_page.id)
        for elt in created_elements:
            self.assertEqual(elt.rotation_angle, 90)

    def test_bulk_create_inherit_mirrored(self):
        """
        Created children elements inherit their parent's mirroring
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:elements-bulk-create", kwargs={"pk": str(self.mirrored_page.id)}),
            data=self.payload,
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_elements = Element.objects.get_descending(self.mirrored_page.id)
        for elt in created_elements:
            self.assertEqual(elt.mirrored, True)

    def test_bulk_create_outside_image(self):
        """
        Cannot create elements outside their image
        """
        self.client.force_login(self.user)
        payload = {
            "worker_run_id": str(self.local_worker_run.id),
            "elements": [
                {
                    "name": "A",
                    "type": "act",
                    "polygon": [
                        [0, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    "name": "B",
                    "type": "surface",
                    # Out of bounds
                    "polygon": [
                        [1003, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    "name": "C",
                    "type": "surface",
                    # Also out of bounds
                    "polygon": [
                        [0, 0],
                        [0, 9],
                        [9, 1200],
                        [9, 0],
                        [0, 0],
                    ]
                }
            ]
        }
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data=payload,
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "elements": {
                "1": {"polygon": ["An element's polygon must not exceed its image's bounds."]},
                "2": {"polygon": ["An element's polygon must not exceed its image's bounds."]}
            }
        })

    def test_bulk_create_unknown_worker_run(self):
        self.client.force_login(self.user)
        random_uuid = str(uuid4())
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                {
                    "worker_run_id": random_uuid,
                    "elements": [
                        {
                            "name": "Blah",
                            "type": "surface",
                            "polygon": [[0, 10], [10, 20], [30, 40], [50, 60], [0, 10]]
                        }
                    ]
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [f'Invalid pk "{random_uuid}" - object does not exist.']
        })

    def test_bulk_create_worker_run_required(self):
        """
        Worker run attribute is required, worker version attribute is forbidden
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data={
                    "elements": [
                        {
                            "name": "Blah",
                            "type": "surface",
                            "polygon": [[0, 10], [10, 20], [30, 40], [50, 60], [0, 10]]
                        }
                    ]
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": ["This field is required."]
        })

    def test_worker_run_non_local(self):
        """
        A regular user cannot create elements with a WorkerRun of a non-local process
        """
        self.client.force_login(self.superuser)
        payload = {**self.payload, "worker_run_id": str(self.worker_run.id)}
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data=payload,
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_run_id": [
            "Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."
        ]})

    def test_worker_run_other_user(self):
        """
        A regular user cannot create elements with a WorkerRun of someone else's local process
        """
        worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)
        payload = {**self.payload, "worker_run_id": str(worker_run.id)}
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data=payload,
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_run_id": [
            "Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."
        ]})

    def test_worker_run_other_process(self):
        """
        A Ponos task cannot create elements with a WorkerRun of another process
        """
        process2 = self.worker_run.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_run.version, parents=[])
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()
        payload = {**self.payload, "worker_run_id": str(other_worker_run.id)}

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data=payload,
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_run_id": [
            "Only the WorkerRuns of the authenticated task's process may be used."
        ]})

    def test_bulk_create_local_worker_run(self):
        """
        A regular user can create elements with a WorkerRun of their own local process
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data=self.payload,
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        elements = (
            Element.objects
            .get_descending(self.element.id)
            .filter(type__slug__in=("act", "surface"))
            .order_by("name")
        )
        element_ids = list(elements.values_list("id", flat=True))
        # Only IDs are returned in the payload
        self.assertListEqual(response.json(), [{"id": str(elt_id)} for elt_id in element_ids])
        # Test common attributes
        parent_path = self.element.paths.get().path
        self.assertListEqual(
            list(elements.values_list(
                "worker_version_id",
                "worker_run_id",
                "image_id",
                "rotation_angle",
                "mirrored",
                "paths__path",
            )),
            3 * [(self.custom_version.id, self.local_worker_run.id, self.element.image_id, 0, False, [*parent_path, self.element.id])]
        )
        # Test specific attributes
        self.assertListEqual(
            list(elements.values_list("name", "type__slug", "paths__ordering", "polygon")),
            [
                ("A", "act", 0, LineString(self.payload["elements"][0]["polygon"])),
                ("B", "surface", 1, LineString(self.payload["elements"][1]["polygon"])),
                ("C", "surface", 2, LineString(self.payload["elements"][2]["polygon"])),
            ],
        )

    def test_bulk_create_task_auth(self):
        """
        Elements can be created with a WorkerRun of a non-local process
        when authenticated as a Ponos task of this process
        """
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()
        payload = {**self.payload, "worker_run_id": str(self.worker_run.id)}

        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data=payload,
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        elements = (
            Element.objects
            .get_descending(self.element.id)
            .filter(type__slug__in=("act", "surface"))
            .order_by("name")
        )
        element_ids = list(elements.values_list("id", flat=True))
        # Ensure elements have been created with valid attributes
        self.assertListEqual(response.json(), [{"id": str(elt_id)} for elt_id in element_ids])
        parent_path = self.element.paths.get().path
        self.assertListEqual(
            list(elements.values_list(
                "worker_version_id",
                "worker_run_id",
                "image_id",
                "rotation_angle",
                "mirrored",
                "paths__path",
            )),
            3 * [(
                self.worker_version.id,
                self.worker_run.id,
                self.element.image_id,
                0,
                False,
                [*parent_path, self.element.id]
            )]
        )

    def test_worker_run_local_task_auth(self):
        """
        Elements can be created with a WorkerRun of a Local process
        even when authenticated as a Ponos task from a different process
        """
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        payload = {**self.payload, "worker_run_id": str(self.local_worker_run.id)}

        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:elements-bulk-create", kwargs={"pk": str(self.element.id)}),
                data=payload,
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        elements = (
            Element.objects
            .get_descending(self.element.id)
            .filter(type__slug__in=("act", "surface"))
            .order_by("name")
        )
        element_ids = list(elements.values_list("id", flat=True))
        # Ensure elements have been created with valid attributes
        self.assertListEqual(response.json(), [{"id": str(elt_id)} for elt_id in element_ids])
        parent_path = self.element.paths.get().path
        self.assertListEqual(
            list(elements.values_list(
                "worker_version_id",
                "worker_run_id",
                "image_id",
                "rotation_angle",
                "mirrored",
                "paths__path",
            )),
            3 * [(
                self.local_worker_run.version_id,
                self.local_worker_run.id,
                self.element.image_id,
                0,
                False,
                [*parent_path, self.element.id]
            )]
        )
