from datetime import datetime, timedelta, timezone
from io import BytesIO
from unittest.mock import call, patch, seal
from uuid import uuid4

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Agent, AgentMode, Farm, State, Task
from arkindex.process.models import Process, ProcessMode
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


@override_settings(PONOS_LOG_TAIL=42)
class TestTaskRestart(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            cls.process.run()
        cls.task1, cls.task2, cls.task3 = cls.process.tasks.all().order_by("depth")

        cls.farm = Farm.objects.first()
        cls.docker_agent = Agent.objects.create(
            mode=AgentMode.Docker,
            farm=cls.farm,
            last_ping=datetime.now(timezone.utc),
            cpu_cores=42,
            cpu_frequency=42e8,
            ram_total=42e3
        )

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(uuid4())})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(uuid4())})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(uuid4())})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.project.mixins.get_max_level")
    def test_forbidden(self, get_max_level_mock):
        """An admin access to the process is required"""
        get_max_level_mock.return_value = Role.Guest.value
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertListEqual(
                response.json(),
                ["You do not have an admin access to the process of this task."],
            )

    def test_non_final_state(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertListEqual(
                response.json(),
                ["Task's state must be in a final state to be restarted."],
            )

    @override_settings(ARKINDEX_FEATURES={"ingest": False})
    def test_s3_requires_feature(self):
        self.process.mode = ProcessMode.S3
        self.process.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertListEqual(
                response.json(),
                ["S3 ingest is not available on this instance."],
            )

    def test_already_restarted(self):
        self.client.force_login(self.user)
        self.task1.slug = self.task1.slug + "_old1"
        self.task1.state = State.Completed.value
        self.task1.save()
        self.task2.original_task_id = self.task1.id
        self.task2.save()
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertListEqual(
                response.json(),
                ["This task has already been restarted."],
            )

    @override_settings(PONOS_RQ_EXECUTION=False)
    @patch("arkindex.project.aws.s3")
    def test_restart_task(self, s3_mock):
        """
        From:
               task2_old1
             ↗
        task1 → task2 → task3
                   ↘      ↗
                    task 4

        To:
               task2_old1
             ↗
        task1  → task2_old2
             ↘
               task2 → task3
                  ↘      ↗
                   task 4
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Task has been restarted")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        task4 = self.process.tasks.create(run=self.task1.run, depth=1)
        task4.parents.add(self.task2)
        task4.children.add(self.task3)
        task_2_slug = self.task2.slug

        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) + timedelta(minutes=1)
            old_task_2 = self.process.tasks.create(run=self.task1.run, depth=1, slug=f"{task_2_slug}_old1")
            old_task_2.state = State.Error.value
            old_task_2.original_task_id = self.task1.id
            old_task_2.save()

        old_task_2.parents.add(self.task1)

        self.task1.state = State.Completed.value
        self.task1.save()
        self.task2.state = State.Error.value
        self.task2.requires_gpu = True

        # Those attributes should not be kept in the new task
        self.task2.started = datetime.now(timezone.utc)
        self.task2.finished = datetime.now(timezone.utc)
        self.task2.agent = self.docker_agent
        self.task2.gpu = self.docker_agent.gpus.create(
            id=uuid4(),
            name="gee pee you",
            index=0,
            ram_total=42e9,
        )
        self.task2.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            with patch("django.utils.timezone.now") as mock_now:
                mock_now.return_value = datetime.now(timezone.utc) + timedelta(minutes=2)
                response = self.client.post(
                    reverse("api:task-restart", kwargs={"pk": str(self.task2.id)})
                )
                self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.process.tasks.count(), 6)
        restarted_task = self.process.tasks.latest("created")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(restarted_task.id),
                "depth": 1,
                "agent": None,
                "extra_files": {},
                "full_log": "http://somewhere",
                "gpu": None,
                "logs": "Task has been restarted",
                "original_task_id": str(self.task2.id),
                "parents": [str(self.task1.id)],
                "run": 0,
                "shm_size": None,
                "slug": task_2_slug,
                "state": "pending",
                "requires_gpu": True,
                "started": None,
                "finished": None,
            },
        )
        self.assertQuerySetEqual(self.task2.children.all(), Task.objects.none())
        self.task2.refresh_from_db()
        self.assertEqual(self.task2.slug, f"{task_2_slug}_old2")
        self.assertNotEqual(self.task2.token, restarted_task.token)
        self.assertIsNone(restarted_task.agent_id)
        self.assertIsNone(restarted_task.gpu_id)
        self.assertQuerySetEqual(
            restarted_task.children.all(),
            Task.objects.filter(id__in=[self.task3.id, task4.id]),
        )

    @override_settings(PONOS_RQ_EXECUTION=False)
    @patch("arkindex.project.aws.s3")
    def test_no_previous_restart(self, s3_mock):
        """
        From:
        task1 → task2 → task3
                   ↘      ↗
                    task 4

        To:
        task1  → task2_old1
             ↘
               task2 → task3
                  ↘      ↗
                   task 4
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Task has been restarted")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        task4 = self.process.tasks.create(run=self.task1.run, depth=1)
        task4.parents.add(self.task2)
        task4.children.add(self.task3)
        task_2_slug = self.task2.slug
        self.task1.state = State.Completed.value
        self.task1.save()
        self.task2.state = State.Error.value
        self.task2.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.process.tasks.count(), 5)
        restarted_task = self.process.tasks.latest("created")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(restarted_task.id),
                "depth": 1,
                "agent": None,
                "extra_files": {},
                "full_log": "http://somewhere",
                "gpu": None,
                "logs": "Task has been restarted",
                "original_task_id": str(self.task2.id),
                "parents": [str(self.task1.id)],
                "run": 0,
                "shm_size": None,
                "slug": task_2_slug,
                "state": "pending",
                "requires_gpu": False,
                "started": None,
                "finished": None,
            },
        )
        self.assertQuerySetEqual(self.task2.children.all(), Task.objects.none())
        self.task2.refresh_from_db()
        self.assertEqual(self.task2.slug, f"{task_2_slug}_old1")
        self.assertQuerySetEqual(
            restarted_task.children.all(),
            Task.objects.filter(id__in=[self.task3.id, task4.id]),
        )

    @override_settings(PONOS_RQ_EXECUTION=False)
    @patch("arkindex.project.aws.s3")
    def test_name_process(self, s3_mock):
        """
        When a task is restarted, its name depends on the existing tasks already existing on the
        same process, not on tasks that exist on other processes
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Task has been restarted")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        task_1_slug = self.task1.slug

        other_process = Process.objects.create(
            farm=Farm.objects.first(),
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.user,
        )
        other_process.tasks.create(
            run=0,
            depth=0,
            slug=f"{task_1_slug}_old1",
            expiry=datetime(1970, 1, 1, tzinfo=timezone.utc),
        )

        self.task1.state = State.Error.value
        self.task1.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            with patch("django.utils.timezone.now") as mock_now:
                mock_now.return_value = datetime.now(timezone.utc) + timedelta(minutes=2)
                response = self.client.post(
                    reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
                )
                self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.process.tasks.count(), 4)
        restarted_task = self.process.tasks.latest("created")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(restarted_task.id),
                "depth": 0,
                "agent": None,
                "extra_files": {},
                "full_log": "http://somewhere",
                "gpu": None,
                "logs": "Task has been restarted",
                "original_task_id": str(self.task1.id),
                "parents": [],
                "run": 0,
                "shm_size": None,
                "slug": task_1_slug,
                "state": "pending",
                "requires_gpu": False,
                "started": None,
                "finished": None,
            },
        )
        self.assertQuerySetEqual(self.task1.children.all(), Task.objects.none())
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.slug, f"{task_1_slug}_old1")
        self.assertNotEqual(self.task1.token, restarted_task.token)
        self.assertIsNone(restarted_task.agent_id)
        self.assertIsNone(restarted_task.gpu_id)
        self.assertQuerySetEqual(
            restarted_task.children.all(),
            Task.objects.filter(id__in=[self.task2.id]),
        )

    @override_settings(PONOS_RQ_EXECUTION=True)
    @patch("arkindex.ponos.tasks.run_task_rq.delay")
    @patch("arkindex.project.aws.s3")
    def test_rq(self, s3_mock, delay_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Task has been restarted")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        delay_mock.return_value = None
        seal(delay_mock)

        self.task1.state = State.Failed.value
        self.task1.save()

        self.client.force_login(self.user)

        with self.assertNumQueries(13), self.captureOnCommitCallbacks(execute=True):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.process.tasks.count(), 4)
        restarted_task = self.process.tasks.latest("created")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(restarted_task.id),
                "depth": 0,
                "agent": None,
                "extra_files": {},
                "full_log": "http://somewhere",
                "gpu": None,
                "logs": "Task has been restarted",
                "original_task_id": str(self.task1.id),
                "parents": [],
                "run": 0,
                "shm_size": None,
                "slug": self.task1.slug,
                "state": "pending",
                "requires_gpu": False,
                "started": None,
                "finished": None,
            },
        )

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(restarted_task))
