from unittest.mock import patch

from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.models import prefetch_related_objects
from django.db.utils import IntegrityError
from django.utils import timezone

from arkindex.ponos.models import FINAL_STATES, Agent, AgentMode, Farm, State
from arkindex.process.models import ProcessMode
from arkindex.project.tests import FixtureAPITestCase


class TestModels(FixtureAPITestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.farm = Farm.objects.create(name="Invisible corn farm")
        cls.process = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
        )
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            cls.process.run()
        cls.task1 = cls.process.tasks.first()
        cls.nonce = b"42" + b"0" * 14

    def test_is_final(self):
        for state in State:
            self.task1.state = state
            self.task1.save()
            if state in FINAL_STATES:
                self.assertTrue(
                    self.task1.is_final(), msg=f"{state} should be final"
                )
                self.assertTrue(
                    self.process.is_final, msg=f"{state} should be final"
                )
            else:
                self.assertFalse(
                    self.task1.is_final(), msg=f"{state} should not be final"
                )
                self.assertFalse(
                    self.process.is_final, msg=f"{state} should not be final"
                )

    def test_requires_gpu(self):
        """
        Check the GPU generated requirements rules
        """
        # Default task does not need GPU
        self.assertFalse(self.task1.requires_gpu)

    @patch("arkindex.ponos.models.timezone")
    def test_task_expiry_default(self, timezone_mock):
        timezone_mock.now.return_value = timezone.datetime(3000, 1, 12).astimezone()
        # Expecting a default expiry 30 days after timezone.now
        expected_expiry = timezone.datetime(3000, 2, 11).astimezone()

        # A process with no tasks has no expiry
        self.process.tasks.all().delete()
        self.assertIsNone(self.process.expiry)

        task1 = self.process.tasks.create(
            command="do something --like this",
            slug="test_task_1",
            run=1,
            depth=0
        )
        task2 = self.process.tasks.create(
            command="do something else --like that",
            slug="test_task_2",
            run=1,
            depth=0
        )

        self.assertEqual(task1.expiry, expected_expiry)
        self.assertEqual(task2.expiry, expected_expiry)
        self.assertEqual(self.process.expiry, expected_expiry)

        # Override a task's expiry
        custom_expiry = timezone.datetime(3000, 4, 20).astimezone()
        task2.expiry = custom_expiry
        task2.save()

        # The workflow's expiry should be the latest expiry
        self.assertEqual(self.process.expiry, custom_expiry)

    def test_process_expiry_query_count(self):
        """
        Workflow.expiry causes an SQL query only when tasks are not prefetched
        """
        with self.assertNumQueries(1):
            self.assertEqual(self.process.expiry, self.task1.expiry)

        # Request the expiry again: it is not cached, there still is an SQL query
        with self.assertNumQueries(1):
            self.assertEqual(self.process.expiry, self.task1.expiry)

        prefetch_related_objects([self.process], "tasks")
        with self.assertNumQueries(0):
            self.assertEqual(self.process.expiry, self.task1.expiry)

    def test_process_get_state(self):
        with self.assertNumQueries(1):
            self.assertEqual(self.process.state, State.Unscheduled)

        with self.assertNumQueries(1):
            self.assertEqual(self.process.get_state(0), State.Unscheduled)

        with self.assertNumQueries(1):
            self.assertEqual(self.process.get_state(1), State.Unscheduled)

        # Negative run numbers should not result in any SQL query, since we know they are always empty
        with self.assertNumQueries(0):
            self.assertEqual(self.process.get_state(-1), State.Unscheduled)

        self.task1.state = State.Running
        self.task1.save()

        with self.assertNumQueries(1):
            self.assertEqual(self.process.state, State.Running)

        with self.assertNumQueries(1):
            self.assertEqual(self.process.get_state(0), State.Running)

        with self.assertNumQueries(1):
            self.assertEqual(self.process.get_state(1), State.Unscheduled)

        with self.assertNumQueries(0):
            self.assertEqual(self.process.get_state(-1), State.Unscheduled)

        prefetch_related_objects([self.process], "tasks")

        with self.assertNumQueries(0):
            self.assertEqual(self.process.state, State.Running)
            self.assertEqual(self.process.get_state(0), State.Running)
            self.assertEqual(self.process.get_state(1), State.Unscheduled)
            self.assertEqual(self.process.get_state(-1), State.Unscheduled)

    def test_agent_default_mode_docker(self):
        Agent.objects.create(
            hostname="agent_smith",
            cpu_cores=2,
            cpu_frequency=4.2e9,
            public_key="",
            farm=self.farm,
            ram_total=2e9,
            last_ping=timezone.now(),
            ram_load=0.49,
            cpu_load=0.99
        )

        test_agent=Agent.objects.get(hostname="agent_smith")
        self.assertEqual(test_agent.mode, AgentMode.Docker)

    def test_agent_slurm_mode(self):
        Agent.objects.create(
            hostname="agent_smith",
            public_key="",
            farm=self.farm,
            last_ping=timezone.now(),
            mode=AgentMode.Slurm.value
        )

        test_agent=Agent.objects.get(hostname="agent_smith")
        self.assertEqual(test_agent.mode, AgentMode.Slurm)

    def test_agent_invalid_mode(self):
        with self.assertRaisesRegex(ValidationError, "bad_mode is not a valid value for enum"):
            Agent.objects.create(
                hostname="agent_smith",
                cpu_cores=2,
                cpu_frequency=4.2e9,
                public_key="",
                farm=self.farm,
                ram_total=2e9,
                last_ping=timezone.now(),
                ram_load=0.49,
                cpu_load=0.99,
                mode="bad_mode"
            )

    def test_agent_mode_hardware_constraint(self):
        """
        When the agent's mode is not AgentMode.Slurm, the hardware requirement parameters
        cannot be null. When the agent's mode is AgentMode.Slurm, they can be null or not.
        """
        cases = [
            {
                # no cpu_cores
                "mode": AgentMode.Docker,
                "params": {
                    "hostname": "agent_smith",
                    "cpu_frequency": 4.2e9,
                    "public_key": "",
                    "farm": self.farm,
                    "ram_total": 2e9,
                    "last_ping": timezone.now(),
                    "ram_load": 0.49,
                    "cpu_load": 0.99,
                },
                "failure": True
            },
            {
                # no cpu_frequency
                "mode": AgentMode.Docker,
                "params": {
                    "hostname": "agent_smith",
                    "cpu_cores": 2,
                    "public_key": "",
                    "farm": self.farm,
                    "ram_total": 2e9,
                    "last_ping": timezone.now(),
                    "ram_load": 0.49,
                    "cpu_load": 0.99,
                },
                "failure": True
            },
            {
                # no ram_total
                "mode": AgentMode.Docker,
                "params": {
                    "hostname": "agent_smith",
                    "cpu_cores": 2,
                    "cpu_frequency": 4.2e9,
                    "public_key": "",
                    "farm": self.farm,
                    "last_ping": timezone.now(),
                    "ram_load": 0.49,
                    "cpu_load": 0.99,
                },
                "failure": True
            },
            {
                # hardware parameters set to None
                "mode": AgentMode.Docker,
                "params": {
                    "hostname": "agent_smith",
                    "cpu_cores": None,
                    "cpu_frequency": None,
                    "public_key": "",
                    "farm": self.farm,
                    "ram_total": None,
                    "last_ping": timezone.now(),
                    "ram_load": 0.49,
                    "cpu_load": 0.99,
                },
                "failure": True
            },
            {
                # slurm mode with hardware requirements
                "mode": AgentMode.Slurm,
                "params": {
                    "hostname": "agent_smith",
                    "cpu_cores": 2,
                    "cpu_frequency": 4.2e9,
                    "public_key": "",
                    "farm": self.farm,
                    "ram_total": 2e9,
                    "last_ping": timezone.now(),
                    "ram_load": 0.49,
                    "cpu_load": 0.99,
                },
                "failure": False
            },
            {
                # slurm mode without hardware requirements
                "mode": AgentMode.Slurm,
                "params": {
                    "hostname": "agent_smith",
                    "public_key": "",
                    "farm": self.farm,
                    "last_ping": timezone.now(),
                    "ram_load": 0.49,
                    "cpu_load": 0.99,
                },
                "failure": False
            },
        ]
        for case in cases:
            with self.subTest(case=case):
                if case["failure"]:
                    # Each create must be run in an atomic block to avoid TransactionManagementError
                    with transaction.atomic():
                        with self.assertRaisesRegex(IntegrityError, "slurm_or_hardware_requirements"):
                            Agent.objects.create(**case["params"], mode=case["mode"])
                else:
                    Agent.objects.create(**case["params"], mode=case["mode"])
