import logging
import operator
from functools import lru_cache, reduce

from django.db import connections
from django.db.models import Exists, Manager, ManyToOneRel, OuterRef, Q
from django.db.models.functions import Coalesce
from django.db.models.query import QuerySet

from arkindex.process.models import Task
from arkindex.users.managers import BaseACLManager
from arkindex.users.models import Role

logger = logging.getLogger(__name__)


class ActivityManager(Manager):
    """Model management for worker activities"""

    def bulk_insert(self, worker_version_id, process_id, configuration_id, model_version_id, elements_qs):
        """
        Create initial worker activities from a queryset of elements in a efficient way.
        Due to the possible large amount of elements, we use a bulk insert from the elements query (best performances).
        The `ON CONFLICT` clause allows either skipping existing activities if they are in a `Started` or `Processed`
        state, or stealing them from another process if they are `Queued` or `Error`.
        """
        sql, params = elements_qs.distinct().values("id").query.sql_with_params()
        select_params = (worker_version_id, configuration_id, model_version_id, process_id) + params

        # With ON CONFLICT, the target constraint is only optional when the action is DO NOTHING.
        # For DO UPDATE, the target can either be a constraint by name with `ON CONSTRAINT <name>`,
        # or the definition of an index that PostgreSQL will look for, using a list of columns,
        # a collation, an operator class, a predicate, etc.
        # `ON CONSTRAINT <name>` does not work here because it is unable to detect partial constraints.
        # We therefore have to explicitly define one of the partial unique constraints ourselves.
        conflict_target_fields = ["worker_version_id", "element_id"]
        if configuration_id is None:
            conflict_target_condition = "configuration_id IS NULL"
        else:
            conflict_target_fields.append("configuration_id")
            conflict_target_condition = "configuration_id IS NOT NULL"

        if model_version_id is None:
            conflict_target_condition += " AND model_version_id IS NULL"
        else:
            conflict_target_fields.append("model_version_id")
            conflict_target_condition += " AND model_version_id IS NOT NULL"

        conflict_target = f'({", ".join(conflict_target_fields)}) WHERE {conflict_target_condition}'

        with connections["default"].cursor() as cursor:
            cursor.execute(
                f"""
                INSERT INTO process_workeractivity
                    (element_id, worker_version_id, configuration_id, model_version_id, state, process_id, id, created, updated)
                SELECT
                    elt.id, %s, %s, %s, 'queued', %s, gen_random_uuid(), current_timestamp, current_timestamp
                FROM ({sql}) AS elt
                ON CONFLICT {conflict_target} DO UPDATE SET
                    process_id = EXCLUDED.process_id,
                    state = EXCLUDED.state,
                    created = EXCLUDED.created
                WHERE process_workeractivity.state <> 'started' AND process_workeractivity.state <> 'processed'
                """,
                select_params)


class CorpusWorkerVersionManager(Manager):

    def rebuild(self):
        """
        Rebuild the corpus worker versions cache from all ML results.
        For compatibility, results without a worker run are generated from the worker version instead.
        """
        from arkindex.documents.models import (
            Classification,
            Element,
            Entity,
            MetaData,
            Transcription,
            TranscriptionEntity,
        )

        querysets = [
            (
                Element.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce("worker_run__version_id", "worker_version_id"))
                .values_list("corpus_id", "version_id", "worker_run__model_version_id", "worker_run__configuration_id")
            ), (
                Transcription.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce("worker_run__version_id", "worker_version_id"))
                .values_list("element__corpus_id", "version_id", "worker_run__model_version_id", "worker_run__configuration_id")
            ), (
                Entity.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce("worker_run__version_id", "worker_version_id"))
                .values_list("corpus_id", "version_id", "worker_run__model_version_id", "worker_run__configuration_id")
            ), (
                TranscriptionEntity.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce("worker_run__version_id", "worker_version_id"))
                .values_list("entity__corpus_id", "version_id", "worker_run__model_version_id", "worker_run__configuration_id")
            ), (
                Classification.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce("worker_run__version_id", "worker_version_id"))
                .values_list("element__corpus_id", "version_id", "worker_run__model_version_id", "worker_run__configuration_id")
            ), (
                MetaData.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce("worker_run__version_id", "worker_version_id"))
                .values_list("element__corpus_id", "version_id", "worker_run__model_version_id", "worker_run__configuration_id")
            ),
        ]

        for i, queryset in enumerate(querysets, start=1):
            logger.info(f"Rebuilding cache from {queryset.model.__name__} ({i}/{len(querysets)})")
            self.bulk_create([
                self.model(
                    corpus_id=corpus_id,
                    worker_version_id=worker_version_id,
                    model_version_id=model_version_id,
                    worker_configuration_id=worker_configuration_id,
                )
                for corpus_id, worker_version_id, model_version_id, worker_configuration_id in queryset.distinct()
            ], ignore_conflicts=True)


class WorkerResultSourceQuerySet(QuerySet):

    @property
    def worker_result_relations(self):
        from arkindex.process.models import WorkerRun
        return (
            field
            for field in self.model._meta.get_fields()
            if isinstance(field, ManyToOneRel)
            # Ignore reverse links to processes tasks
            and field.related_model is not Task
            # Ignore the WorkerRun→WorkerVersion link
            and field.related_model is not WorkerRun
        )

    def in_use(self):
        """
        Check if any data is linked to this worker run or worker version
        """
        ids = self.values_list("id", flat=True)
        return any(
            relation.related_model.objects.filter(**{f"{relation.field.name}__in": ids}).exists()
            for relation in self.worker_result_relations
        )

    def get_by_feature(self, feature):
        try:
            return self.select_related("worker").get(feature=feature)
        except self.model.DoesNotExist:
            raise self.model.DoesNotExist(f"There are no worker versions supporting the {feature.name} feature.")


class WorkerVersionManager(Manager):

    @lru_cache
    def get_by_feature(self, feature):
        return self.get_queryset().get_by_feature(feature)

    def get_queryset(self):
        return WorkerResultSourceQuerySet(self.model, using=self._db)


class WorkerManager(BaseACLManager):

    def executable(self, user):
        """
        Workers that can be viewed and executed by a user.
        """
        # Without authentication, or as a Ponos agent, nothing is executable
        # Agents are not users, filtering on them will make Django raise exceptions, so we have to handle them separately.
        if user.is_anonymous or getattr(user, "is_agent", False):
            return self.none()

        # Admins can execute anything
        if user.is_admin:
            return self.all()

        return self.filter(
            # All public workers are executable by everyone
            Q(public=True)
            # Otherwise, require contributor rights on either the worker
            | Q(id__in=self.filter_rights(user, self.model, Role.Contributor.value).values("id"))
        )


class WorkerRunQuerySet(WorkerResultSourceQuerySet):

    def set_has_results(self):
        """
        Set the has_results field on worker runs to True if these workers runs are linked to existing machine learning results
        """
        queryset = list(self.annotate(new_has_results=self.has_results_expression()).only("id"))
        for run in queryset:
            run.has_results = run.new_has_results
        self.bulk_update(queryset, ["has_results"], 1000)

    def has_results_expression(self):
        return reduce(operator.or_, [
            Exists(
                relation.related_model.objects.filter(worker_run_id=OuterRef("pk"))
            )
            for relation in self.worker_result_relations
        ])


class WorkerRunManager(Manager):
    def get_queryset(self):
        return WorkerRunQuerySet(self.model, using=self._db)
