import urllib
from typing import Optional

from django.conf import settings
from django.core.mail import send_mail
from django.db import transaction
from django.template.loader import render_to_string
from django.urls import reverse
from django_rq import job
from rq import Retry, get_current_job

from arkindex.process.models import (
    ActivityState,
    Process,
    ProcessMode,
    WorkerActivity,
    WorkerActivityState,
)
from arkindex.users.models import User


@job("default", timeout=settings.RQ_TIMEOUTS["initialize_activity"], retry=Retry(max=4))
def initialize_activity(process: Process):
    """
    List all worker versions used in a process and initialize their activity on processed elements.
    4 retries allowed, for a total of 5 attempts, to try to mitigate some database errors from the large query.
    """
    worker_runs = [
        worker_run for worker_run in (
            process.worker_runs
            .select_related("version__worker")
            .only("process_id", "version_id", "configuration_id", "model_version_id", "version__worker__slug")
        )
        # Do not generate worker activities for the elements initialisation task
        if not worker_run.version.is_init_elements()
    ]
    rq_job = get_current_job()

    try:
        with transaction.atomic():
            for i, worker_run in enumerate(worker_runs):
                if rq_job is not None:
                    rq_job.set_progress(i / len(worker_runs))
                WorkerActivity.objects.bulk_insert(
                    worker_version_id=worker_run.version_id,
                    process_id=process.id,
                    configuration_id=worker_run.configuration_id,
                    model_version_id=worker_run.model_version_id,
                    elements_qs=process.list_elements()
                )
    except Exception as e:
        process.activity_state = ActivityState.Error
        process.save()
        raise e
    process.activity_state = ActivityState.Ready
    process.save()


@job("default", timeout=settings.RQ_TIMEOUTS["process_delete"])
def process_delete(process: Process, delete_activity: Optional[WorkerActivityState] = None):
    """
    Delete a process and removes its reference from related worker activities.
    If `delete` is defined, activities of this state will be deleted.
    """
    if delete_activity:
        process.activities.filter(state=delete_activity).delete()
    process.delete()


@job("default", timeout=settings.RQ_TIMEOUTS["create_process_failures"])
def create_process_failures(base_process: Process):
    """
    Create a process containing elements from previous WorkerActivity failures.
    Once it is done, warn the user by email with a link to process' configuration.
    """
    rq_job = get_current_job()
    assert rq_job is not None, "This task can only be run in a RQ job context."
    assert rq_job.user_id is not None, "This task requires a user ID to be defined on the RQ job."
    user = User.objects.get(id=rq_job.user_id)

    with transaction.atomic():
        process = Process.objects.create(
            corpus=base_process.corpus,
            mode=ProcessMode.Workers,
            name=f"Failures from {base_process.id}",
            creator=user,
        )
        process.elements.set(
            base_process.activities
            .filter(state=WorkerActivityState.Error)
            .distinct("element_id")
            .values_list("element_id", flat=True)
        )
    process_edit_url = urllib.parse.urljoin(
        settings.PUBLIC_HOSTNAME,
        reverse("frontend-configure-process", kwargs={"pk": process.id}),
    )
    subject="Process created from previous failures"
    if base_process.name:
        subject += f" ({base_process.name})"
    send_mail(
        subject=subject,
        message=render_to_string("process_failures.html", context={"user": user, "url": process_edit_url}),
        from_email=None,
        recipient_list=[user.email],
        fail_silently=False,
    )
