from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import Worker, WorkerType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestWorkerTypes(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_type_dla = WorkerType.objects.get(slug="dla")
        cls.worker_type_recognizer = WorkerType.objects.get(slug="recognizer")
        cls.worker_type_worker = WorkerType.objects.get(slug="worker")
        cls.worker_type_import = WorkerType.objects.get(slug="import")
        cls.worker_type_custom = WorkerType.objects.get(slug="custom")
        cls.init_type = WorkerType.objects.get(slug="init")

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:worker-type-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_list_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:worker-type-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.utils.filter_rights")
    def test_list_admin(self, filter_rights_mock):
        """
        Admins can see every type
        """
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:worker-type-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 6,
            "previous": None,
            "next": None,
            "number": 1,
            "results": [
                {
                    "display_name": "Custom",
                    "id": str(self.worker_type_custom.id),
                    "slug": "custom",
                },
                {
                    "id": str(self.worker_type_dla.id),
                    "slug": "dla",
                    "display_name": "Document Layout Analysis",
                },
                {
                    "id": str(self.init_type.id),
                    "slug": "init",
                    "display_name": "Elements Initialisation",
                },
                {
                    "id": str(self.worker_type_import.id),
                    "slug": "import",
                    "display_name": "Import",
                },
                {
                    "id": str(self.worker_type_recognizer.id),
                    "slug": "recognizer",
                    "display_name": "Recognizer",
                },
                {
                    "id": str(self.worker_type_worker.id),
                    "slug": "worker",
                    "display_name": "Worker requiring a GPU",
                },
            ],
        })

        # Rights are not checked
        self.assertFalse(filter_rights_mock.called)

    @patch("arkindex.users.utils.filter_rights")
    def test_list_user(self, filter_rights_mock):
        """
        Regular users can only see the types with workers that they have access to
        """
        self.client.force_login(self.user)

        # Emulate a contributor access to a `dla` worker
        self.assertTrue(self.worker_type_dla.workers.exists())
        filter_rights_mock.return_value = Worker.objects.filter(type=self.worker_type_dla)
        # The init elements and file import workers are public
        self.assertTrue(self.init_type.workers.exists())
        self.assertTrue(self.worker_type_import.workers.exists())

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:worker-type-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 3,
            "previous": None,
            "next": None,
            "number": 1,
            "results": [
                {
                    "id": str(self.worker_type_dla.id),
                    "slug": "dla",
                    "display_name": "Document Layout Analysis",
                },
                {
                    "id": str(self.init_type.id),
                    "slug": "init",
                    "display_name": "Elements Initialisation",
                },
                {
                    "id": str(self.worker_type_import.id),
                    "slug": "import",
                    "display_name": "Import",
                }
            ],
        })

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Worker, Role.Contributor.value))
