from datetime import datetime, timezone

from arkindex.documents.models import Element
from arkindex.ponos.models import Farm
from arkindex.process.models import (
    Process,
    ProcessMode,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerVersion,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState


class TestWorkerActivityBulkInsert(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            corpus=cls.corpus,
            farm=Farm.objects.first(),
        )
        cls.configuration = WorkerConfiguration.objects.create(worker=cls.worker_version.worker, name="A config", configuration={"a": "b"})
        cls.model = Model.objects.create(name="Mochi", public=False)
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            hash="a",
        )
        cls.worker_run = cls.process.worker_runs.create(
            version=cls.worker_version,
            configuration=cls.configuration,
            model_version=cls.model_version,
        )

    def test_worker_version(self):
        """
        Bulk insert worker activities for acts
        """
        activities = WorkerActivity.objects.filter(
            state=WorkerActivityState.Queued,
            process=self.process,
            configuration=None,
            model_version=None,
        )
        self.assertEqual(activities.count(), 0)

        elements_qs = Element.objects.filter(type__slug="act", type__corpus_id=self.corpus.id)
        params = {
            "worker_version_id": self.worker_version.id,
            "corpus_id": self.corpus.id,
            "process_id": self.process.id,
        }
        with self.assertExactQueries("workeractivity_bulk_insert_worker_version_only.sql", params=params):
            WorkerActivity.objects.bulk_insert(
                worker_version_id=self.worker_version.id,
                process_id=self.process.id,
                configuration_id=None,
                model_version_id=None,
                elements_qs=elements_qs,
            )

        self.assertEqual(elements_qs.count(), 5)
        self.assertEqual(activities.count(), 5)

    def test_no_model(self):
        """
        Bulk insert worker activities for acts
        """
        activities = WorkerActivity.objects.filter(
            state=WorkerActivityState.Queued,
            process=self.process,
            configuration=self.configuration,
            model_version=None,
        )
        self.assertEqual(activities.count(), 0)

        elements_qs = Element.objects.filter(type__slug="act", type__corpus_id=self.corpus.id)
        params = {
            "worker_version_id": self.worker_version.id,
            "corpus_id": self.corpus.id,
            "process_id": self.process.id,
            "configuration_id": self.configuration.id,
        }
        with self.assertExactQueries("workeractivity_bulk_insert_no_model.sql", params=params):
            WorkerActivity.objects.bulk_insert(
                worker_version_id=self.worker_version.id,
                process_id=self.process.id,
                configuration_id=self.configuration.id,
                model_version_id=None,
                elements_qs=elements_qs,
            )

        self.assertEqual(elements_qs.count(), 5)
        self.assertEqual(activities.count(), 5)

    def test_no_configuration(self):
        """
        Bulk insert worker activities for acts
        """
        activities = WorkerActivity.objects.filter(
            state=WorkerActivityState.Queued,
            process=self.process,
            configuration=None,
            model_version=self.model_version,
        )
        self.assertEqual(activities.count(), 0)

        elements_qs = Element.objects.filter(type__slug="act", type__corpus_id=self.corpus.id)
        params = {
            "worker_version_id": self.worker_version.id,
            "model_version_id": self.model_version.id,
            "corpus_id": self.corpus.id,
            "process_id": self.process.id,
        }
        with self.assertExactQueries("workeractivity_bulk_insert_no_configuration.sql", params=params):
            WorkerActivity.objects.bulk_insert(
                worker_version_id=self.worker_version.id,
                process_id=self.process.id,
                configuration_id=None,
                model_version_id=self.model_version.id,
                elements_qs=elements_qs,
            )

        self.assertEqual(elements_qs.count(), 5)
        self.assertEqual(activities.count(), 5)

    def test_duplicate_elements(self):
        """
        WorkerActivity.bulk_insert should exclude duplicate elements
        """
        element_type = self.corpus.types.first()
        parent1 = self.corpus.elements.create(type=element_type, name="Parent 1")
        parent2 = self.corpus.elements.create(type=element_type, name="Parent 2")
        element = self.corpus.elements.create(type=element_type, name="Element")
        child = self.corpus.elements.create(type=element_type, name="Child")
        element.add_parent(parent1)
        element.add_parent(parent2)
        child.add_parent(element)
        elements_qs = Element.objects.filter(paths__path__contains=[element.id], name="Child")
        # `child` has two paths that both contain the ID of `element`, because `element` has two parents,
        # so filtering on paths__path will duplicate the child
        self.assertEqual(elements_qs.count(), 2)

        WorkerActivity.objects.bulk_insert(
            worker_version_id=self.worker_version.id,
            process_id=self.process.id,
            configuration_id=self.configuration.id,
            model_version_id=self.model_version.id,
            elements_qs=elements_qs,
        )
        self.assertEqual(self.process.activities.get().element_id, child.id)

    def test_bulk_insert(self):
        """
        Bulk insert worker activities for acts
        """
        # Create an activity on another element that should not be erased
        self.corpus.elements.get(name="Volume 1, page 2r").activities.create(
            process=self.process,
            worker_version=self.worker_version,
            state=WorkerActivityState.Queued,
            configuration=self.configuration,
            model_version=self.model_version,
        )

        activities = WorkerActivity.objects.filter(
            state=WorkerActivityState.Queued,
            process=self.process,
            configuration=self.configuration,
            model_version=self.model_version,
        )
        self.assertEqual(activities.count(), 1)

        elements_qs = Element.objects.filter(type__slug="act", type__corpus_id=self.corpus.id)
        params = {
            "worker_version_id": self.worker_version.id,
            "configuration_id": self.configuration.id,
            "model_version_id": self.model_version.id,
            "corpus_id": self.corpus.id,
            "process_id": self.process.id,
        }
        with self.assertExactQueries("workeractivity_bulk_insert.sql", params=params):
            WorkerActivity.objects.bulk_insert(
                worker_version_id=self.worker_version.id,
                process_id=self.process.id,
                configuration_id=self.configuration.id,
                model_version_id=self.model_version.id,
                elements_qs=elements_qs,
            )

        self.assertEqual(elements_qs.count(), 5)
        # All acts + the existing activity
        self.assertEqual(activities.count(), 6)

    def test_existing(self):
        """
        WorkerActivity.objects.bulk_insert should steal existing WorkerActivities from other processes
        if those are in a queued or error state
        """
        old_process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
        )
        elements_qs = Element.objects.filter(type__slug="act", type__corpus_id=self.corpus.id)
        self.assertEqual(elements_qs.count(), 5)

        self.process.activities.all().delete()

        # Create one WorkerActivity for each possible state
        WorkerActivity.objects.bulk_create([
            WorkerActivity(
                element=element,
                worker_version=self.worker_version,
                state=state,
                process=old_process,
                configuration=self.configuration,
                model_version=self.model_version,
                started=datetime.now(timezone.utc),
            )
            for element, state in zip(elements_qs, WorkerActivityState)
        ])

        self.assertEqual(self.process.activities.count(), 0)
        self.assertEqual(old_process.activities.count(), 4)
        self.assertCountEqual(list(old_process.activities.values_list("state", flat=True).distinct()), [
            WorkerActivityState.Queued,
            WorkerActivityState.Started,
            WorkerActivityState.Processed,
            WorkerActivityState.Error,
        ])

        with self.assertNumQueries(1):
            WorkerActivity.objects.bulk_insert(
                worker_version_id=self.worker_version.id,
                process_id=self.process.id,
                configuration_id=self.configuration.id,
                model_version_id=self.model_version.id,
                elements_qs=elements_qs,
            )

        # The old process kept the two started and processed activities
        self.assertEqual(old_process.activities.count(), 2)
        self.assertCountEqual(list(old_process.activities.values_list("state", flat=True).distinct()), [
            WorkerActivityState.Started,
            WorkerActivityState.Processed,
        ])

        # The new process gets the two queued and error activities in a started state as requested,
        # and also gets one extra activity for the act that had no activity before
        self.assertEqual(self.process.activities.count(), 3)
        self.assertCountEqual(list(self.process.activities.values_list("state", flat=True).distinct()), [
            WorkerActivityState.Queued,
        ])
