from datetime import datetime, timezone
from unittest.mock import patch
from uuid import uuid4

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Element
from arkindex.ponos.models import Farm
from arkindex.process.models import (
    Process,
    ProcessMode,
    Worker,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState


@override_settings(PUBLIC_HOSTNAME="https://darkindex.lol")
class TestUpdateWorkerActivity(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.element = Element.objects.get(name="Volume 1, page 2r")
        cls.process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            corpus=cls.corpus,
            farm=Farm.objects.first(),
        )
        cls.configuration = WorkerConfiguration.objects.create(worker=cls.worker_version.worker, name="A config", configuration={"a": "b"})
        cls.model = Model.objects.create(name="Mochi", public=False)
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            hash="a",
        )
        cls.worker_run = cls.process.worker_runs.create(
            version=cls.worker_version,
            configuration=cls.configuration,
            model_version=cls.model_version,
        )

        # Run the process, but skip the real activity initialization so that we can control it ourselves
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            cls.process.run()

        cls.task = cls.process.tasks.get(slug=cls.worker_run.task_slug)

        # Create a queued activity for this element
        cls.activity = cls.element.activities.create(
            process=cls.process,
            worker_version=cls.worker_version,
            state=WorkerActivityState.Queued,
            configuration=cls.configuration,
            model_version=cls.model_version,
        )

    def test_user(self):
        """
        Only Ponos task authentication is allowed on UpdateWorkerActivity
        """
        for user in [None, self.user, self.superuser]:
            self.activity.state = WorkerActivityState.Queued
            self.activity.save(update_fields=["state"])
            if user:
                self.client.force_login(user)
            else:
                self.client.logout()
            with self.assertNumQueries(0):
                response = self.client.put(
                    reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                    {
                        "element_id": str(self.element.id),
                        "process_id": str(self.process.id),
                        "state": WorkerActivityState.Started.value,
                    },
                )
                self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_task(self):
        """
        Ponos task authentication can be used to update a WorkerActivity.
        Tasks can update any WorkerActivity in their process, without any filter on the WorkerVersion,
        since there is no link between a WorkerVersion and a Task yet.
        """

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                {
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    def test_task_other_process(self):
        """
        Updating a WorkerActivity of another process should be forbidden.
        Unlike the usual HTTP 409, we can raise a HTTP 400 here as the check requires no extra query and can be done
        in the serializer before the UPDATE query that normally updates any WorkerActivity if it exists.
        """
        process2 = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
            farm=Farm.objects.first(),
        )
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            process2.run()
        task = process2.tasks.first()

        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                {
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "process_id": ["Only WorkerActivities for the process of the currently authenticated task can be updated."],
        })

        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Queued)

    def test_wrong_worker_version(self):
        """
        Raises a generic error in case the worker version does not exist because a single SQL request is performed
        The response is a HTTP_409_CONFLICT
        """
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(uuid4())}),
                {
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            "__all__": [
                "Either this worker activity does not exist, is assigned to another process, or "
                f"updating the state to '{WorkerActivityState.Started.value}' is forbidden."
            ]
        })

    def test_not_found_no_process_id(self):
        """
        When a WorkerRun does not exist for this ID, and there is no process_id, we cannot find a WorkerVersion either
        """
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(uuid4())}),
                {
                    "element_id": str(self.element.id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "detail": [
                "Either this WorkerRun ID does not exist, or a WorkerVersion ID was sent without a `process_id`.",
            ]
        })

    def test_multiple_worker_versions(self):
        """
        Calling UpdateWorkerActivity with a worker version ID is not supported
        when the same version is used multiple times in the same process
        """
        # Create a second run with the same version
        self.process.worker_runs.create(
            version=self.worker_version,
            parents=[self.worker_run.id],
            # Different configuration
            configuration=None,
            model_version=self.model_version,
        )

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                {
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "detail": [
                "There is more than one WorkerRun on this process with this WorkerVersion. "
                "Please update the worker to call UpdateWorkerActivity with WorkerRun IDs, "
                "run the duplicated WorkerVersion in a process without worker activities, "
                "or run each duplicated WorkerVersion in a separate process."
            ]
        })

    def test_element_unexisting(self):
        """
        Raises a generic error in case no activity exists for this element
        The response is a HTTP_409_CONFLICT
        """
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                {
                    "element_id": str(uuid4()),
                    "process_id": str(self.process.id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            "__all__": [
                "Either this worker activity does not exist, is assigned to another process, or "
                f"updating the state to '{WorkerActivityState.Started.value}' is forbidden."
            ]
        })

    def test_process_unexisting(self):
        """
        Raises an error in case the process does not exist
        """
        wrong_process_id = uuid4()
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                {
                    "element_id": str(self.element.id),
                    "process_id": str(wrong_process_id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "process_id": [f'Invalid pk "{wrong_process_id}" - object does not exist.']
        })

    def test_worker_run(self):
        """
        UpdateWorkerActivity can be called with a WorkerRun ID and no process ID
        """
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_run.id)}),
                {
                    "element_id": str(self.element.id),
                    "state": "started",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    def test_allowed_states(self):
        """
        Check each case state update is allowed depending on the actual state of the activity
        """
        queued, started, error, processed = (
            WorkerActivityState[state].value
            for state in ["Queued", "Started", "Error", "Processed"]
        )
        allowed_states_update = (
            (self.worker_version.id, queued, started, 5),
            (self.worker_version.id, started, processed, 5),
            (self.worker_version.id, started, error, 6),
            (self.worker_version.id, error, queued, 5),
            (self.worker_version.id, error, started, 5),
            (self.worker_run.id, queued, started, 4),
            (self.worker_run.id, started, processed, 4),
            (self.worker_run.id, started, error, 5),
            (self.worker_run.id, error, queued, 4),
            (self.worker_run.id, error, started, 4),
        )
        known_date = datetime(2012, 12, 12, 12, 12, 12, 12, timezone.utc)
        for version_or_run_id, state, payload_state, nb_query in allowed_states_update:
            with self.subTest(previous_state=state, new_state=payload_state):
                self.activity.state = state
                if state == started:
                    self.activity.started = known_date
                else:
                    self.activity.started = None
                self.activity.save(update_fields=["state", "started"])

                with self.assertNumQueries(nb_query):
                    response = self.client.put(
                        reverse("api:update-worker-activity", kwargs={"pk": str(version_or_run_id)}),
                        {
                            "element_id": str(self.element.id),
                            "process_id": str(self.process.id),
                            "state": payload_state,
                        },
                        HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.activity.refresh_from_db()
                self.assertEqual(self.activity.state.value, payload_state)
                if state == started:
                    # When updating from started to another state, the started time should not be changed at all
                    self.assertEqual(self.activity.started, known_date)
                elif payload_state == started:
                    # When updating to started, we had set a None start timestamp, so it shouldn't be None anymore
                    self.assertIsNotNone(self.activity.started)
                else:
                    # When the started state is not involved, the start timestamp is not set
                    self.assertIsNone(self.activity.started)

    def test_forbidden_states(self):
        """
        Check state update is forbidden for some non consistent cases
        The response is a HTTP_409_CONFLICT
        """
        queued, started, error, processed = (
            WorkerActivityState[state].value
            for state in ["Queued", "Started", "Error", "Processed"]
        )
        forbidden_states_update = (
            (self.worker_version.id, queued, queued, 5),
            (self.worker_version.id, started, queued, 5),
            (self.worker_version.id, queued, error, 5),
            (self.worker_version.id, processed, error, 5),
            (self.worker_run.id, queued, queued, 4),
            (self.worker_run.id, started, queued, 4),
            (self.worker_run.id, queued, error, 4),
            (self.worker_run.id, processed, error, 4),
        )
        for version_or_run_id, state, payload_state, query_count in forbidden_states_update:
            self.activity.state = state
            if state == started:
                self.activity.started = datetime.now(timezone.utc)
            else:
                self.activity.started = None
            self.activity.save(update_fields=["state", "started"])
            with self.assertNumQueries(query_count):
                response = self.client.put(
                    reverse("api:update-worker-activity", kwargs={"pk": str(version_or_run_id)}),
                    {
                        "element_id": str(self.element.id),
                        "process_id": str(self.process.id),
                        "state": payload_state,
                    },
                    HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                )
                self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
            self.assertDictEqual(response.json(), {
                "__all__": [
                    "Either this worker activity does not exist, is assigned to another process, or "
                    f"updating the state to '{payload_state}' is forbidden."
                ]
            })
            self.activity.refresh_from_db()
            self.assertEqual(self.activity.state.value, state)

    # Just set the timeout to zero so that the transition always works
    @override_settings(WORKER_ACTIVITY_TIMEOUT=0)
    def test_retry(self):
        self.activity.state = WorkerActivityState.Started
        self.activity.started = datetime.now(timezone.utc)
        self.activity.save(update_fields=["state", "started"])

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                {
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    @override_settings(WORKER_ACTIVITY_TIMEOUT=10000)
    def test_retry_limit(self):
        self.activity.state = WorkerActivityState.Started
        self.activity.started = datetime.now(timezone.utc)
        self.activity.save(update_fields=["state", "started"])

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                {
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "state": WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)

        self.assertDictEqual(response.json(), {
            "__all__": [
                "Either this worker activity does not exist, is assigned to another process, or "
                "updating the state to 'started' is forbidden."
            ]
        })
        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    def test_error_updates_other_activities(self):
        """
        The error update should update the status on the following tasks
        """
        worker_version_2 = WorkerVersion.objects.get(worker__slug="dla")
        worker_version_3 = WorkerVersion.objects.get(worker__slug="worker-gpu")
        worker_version_4 = WorkerVersion.objects.create(
            worker=Worker.objects.create(
                name="New version",
                slug="new",
                type=self.worker_version.worker.type,
            ),
            version=4,
            configuration={},
            state=WorkerVersionState.Available,
            docker_image_iid=self.worker_version.docker_image_iid
        )
        # Create runs run_1 > run_2 > run_3
        run_1 = self.process.worker_runs.first()
        run_2 = self.process.worker_runs.create(
            version=worker_version_2,
            parents=[run_1.id],
        )
        self.process.worker_runs.create(
            version=worker_version_3,
            parents=[run_2.id],
        )

        self.process.worker_runs.create(
            version=worker_version_4,
            parents=[],
        )

        # Create activities for run_2, run_3 and run_4
        activity_2 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_2,
            state=WorkerActivityState.Processed
        )
        activity_3 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_3,
            state=WorkerActivityState.Queued
        )
        activity_4 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_4,
            state=WorkerActivityState.Queued
        )

        self.activity.state = WorkerActivityState.Started
        self.activity.started = datetime.now(timezone.utc)
        self.activity.save(update_fields=["state", "started"])

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:update-worker-activity", kwargs={"pk": str(self.worker_version.id)}),
                {
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "state": WorkerActivityState.Error.value,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Error)
        activity_2.refresh_from_db()
        self.assertEqual(activity_2.state, WorkerActivityState.Processed)
        activity_3.refresh_from_db()
        self.assertEqual(activity_3.state, WorkerActivityState.Error)
        activity_4.refresh_from_db()
        self.assertEqual(activity_4.state, WorkerActivityState.Error)
