from unittest.mock import call, patch
from uuid import uuid4

from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.process.models import Worker
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Group, Right, Role, User


class TestModelAPI(FixtureAPITestCase):
    """
    Test model and model version api
    """

    @classmethod
    def setUpTestData(cls):
        r"""We use a simple rights configuration for those tests

            User1     User2       User3
              |        /           /
              |      100          /
              |       |          /
             100   Group1      10
              |     / \        /
              |   50   100    /
              |  /       \   /
            Model1      Model2
        """
        super().setUpTestData()

        cls.worker = Worker.objects.get(slug="generic")

        # Create three models
        cls.model1 = Model.objects.create(name="First Model", description="first")
        cls.model2 = Model.objects.create(name="Second Model")
        cls.model2.compatible_workers.set([cls.worker])
        cls.model3 = Model.objects.create(name="Third Model")

        # Create some Model Versions
        cls.model_version1 = cls.model1.versions.create(model=cls.model1)
        cls.model_version2 = cls.model1.versions.create(
            description="some description",
            tag="tagged",
            configuration={"n_epochs": 10},
        )
        cls.model_version3 = cls.model2.versions.create(
            state=ModelVersionState.Available,
            tag="tagged",
            hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbba",
            archive_hash="a501b1ae6f6bc833551245f8328590b8",
            size=8,
        )
        # This version has a blank tag but has been validated
        cls.model_version4 = cls.model2.versions.create(
            description="some description",
            tag=None,
            state=ModelVersionState.Available,
            configuration={"n_epochs": 10},
            hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbaa",
            archive_hash="d67459a391b228cced507f068d4a570a",
            size=8,
        )
        cls.model_version5 = cls.model1.versions.create(
            description="some description",
            tag="available_version",
            state=ModelVersionState.Available,
            configuration={"n_epochs": 10},
            hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbdd",
            archive_hash="3d87fd18e03fcbbfb64a381ad9472596",
            size=8,
        )

        # Create three users with different access rights on each model
        cls.user1 = User.objects.create(email="user1@test.test", display_name="User 1", verified_email=True)
        cls.user2 = User.objects.create(email="user2@test.test", display_name="User 2", verified_email=True)
        cls.user3 = User.objects.create(email="user3@test.test", display_name="User 3", verified_email=True)

        # Create the group
        cls.group1 = Group.objects.create(name="Group1")

        # Access rights
        Right.objects.bulk_create([
            Right(user=cls.user1, content_object=cls.model1, level=100),
            Right(user=cls.user2, content_object=cls.group1, level=100),
            Right(user=cls.user3, content_object=cls.model2, level=10),
            Right(group=cls.group1, content_object=cls.model1, level=50),
            Right(group=cls.group1, content_object=cls.model2, level=100),
        ])

       # CreateModel

    def test_create_model_requires_verified(self):
        """Model creation requires being verified
        """
        user = User.objects.create(display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        response = self.client.post(reverse("api:models"), {"name": "The Best Classifier"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_create_model_requires_logged_in(self):
        """Model creation requires being logged in
        """
        response = self.client.post(reverse("api:models"), {"name": "The Best Classifier"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_create_model_requires_name(self):
        """To create a model you must provide a name
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:models"))

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field is required."]})

    def test_create_model_requires_name_not_blank(self):
        """To create a model you must provide a name
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:models"), {"name": ""})

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field may not be blank."]})

    def test_create_model_new_name(self):
        """Create a model with a name no other model has and a description. The creator also gets admin rights on the newly created model.
        """
        self.client.force_login(self.user1)
        request = {
            "name": "The Best Model Ever",
            "description": "This is actually the best model ever."
        }
        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:models"), request)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["rights"], ["read", "write", "admin"])

        model = Model.objects.get(id=response.json()["id"])
        self.assertEqual(model.name, request.get("name"))
        self.assertEqual(model.description, request.get("description"))

        # Check that user has admin rights on the model
        user_right = model.memberships.get(user=self.user1)
        self.assertEqual(user_right.level, Role.Admin.value)

    def test_create_model_name_taken_with_access(self):
        """Raises a 400 with the model_id when creating a model with a name that is already used for another model
        but with guest access rights on this model
        """
        self.client.force_login(self.user1)
        request = {
            "name": self.model1.name,
        }
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:models"), request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"id": str(self.model1.id), "name": "A model with this name already exists"})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_create_model_name_taken_no_access(self, has_access_mock):
        """
        Raises a 403 with no additional information when creating a model with a name that is already used for another model
        but without access rights on this model
        """
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:models"), {
                "name": self.model1.name,
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})
        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user3, self.model1, Role.Guest.value, skip_public=False))

    # ListModels

    def test_list_models_requires_verified(self):
        user = User.objects.create(display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        response = self.client.get(reverse("api:models"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_list_models_requires_logged_in(self):
        response = self.client.get(reverse("api:models"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_list(self):
        self.client.force_login(self.user2)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:models"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        models = response.json()["results"]

        self.assertListEqual(models, [
            {
                "id": str(self.model1.id),
                "created": self.model1.created.isoformat().replace("+00:00", "Z"),
                "updated": self.model1.updated.isoformat().replace("+00:00", "Z"),
                "name": "First Model",
                "description": "first",
                "rights": ["read", "write", "admin"],
                "archived": False,
            },
            {
                "id": str(self.model2.id),
                "created": self.model2.created.isoformat().replace("+00:00", "Z"),
                "updated": self.model2.updated.isoformat().replace("+00:00", "Z"),
                "name": "Second Model",
                "description": "",
                "rights": ["read", "write", "admin"],
                "archived": False,
            },
            {
                "id": str(self.model3.id),
                "created": self.model3.created.isoformat().replace("+00:00", "Z"),
                "updated": self.model3.updated.isoformat().replace("+00:00", "Z"),
                "name": "Third Model",
                "description": "",
                "rights": ["read", "write", "admin"],
                "archived": False,
            },
        ])

    def test_list_models_filter_name(self):
        """User 2 has access to both Models, use search parameter name=Second returns only Model2
        """
        self.client.force_login(self.user2)
        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:models"), {"name": "second"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        models = response.json()["results"]

        self.assertListEqual(models, [
            {
                "id": str(self.model2.id),
                "created": self.model2.created.isoformat().replace("+00:00", "Z"),
                "updated": self.model2.updated.isoformat().replace("+00:00", "Z"),
                "name": "Second Model",
                "description": "",
                "rights": ["read", "write", "admin"],
                "archived": False,
            }
        ])

    def test_list_models_filter_compatible_worker(self):
        self.client.force_login(self.user2)
        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:models"), {"compatible_worker": str(self.worker.id)})
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        models = response.json()["results"]

        self.assertListEqual(models, [
            {
                "id": str(self.model2.id),
                "created": self.model2.created.isoformat().replace("+00:00", "Z"),
                "updated": self.model2.updated.isoformat().replace("+00:00", "Z"),
                "name": "Second Model",
                "description": "",
                "rights": ["read", "write", "admin"],
                "archived": False,
            }
        ])

    def test_list_models_filter_archived(self):
        self.client.force_login(self.user2)
        self.model1.archived = timezone.now()
        self.model1.save()

        cases = [
            (True, [
                {
                    "id": str(self.model1.id),
                    "created": self.model1.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.model1.updated.isoformat().replace("+00:00", "Z"),
                    "name": "First Model",
                    "description": "first",
                    "rights": ["read", "write", "admin"],
                    "archived": True,
                }
            ]),
            (False, [
                {
                    "id": str(self.model2.id),
                    "created": self.model2.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.model2.updated.isoformat().replace("+00:00", "Z"),
                    "name": "Second Model",
                    "description": "",
                    "rights": ["read", "write", "admin"],
                    "archived": False,
                },
                {
                    "id": str(self.model3.id),
                    "created": self.model3.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.model3.updated.isoformat().replace("+00:00", "Z"),
                    "name": "Third Model",
                    "description": "",
                    "rights": ["read", "write", "admin"],
                    "archived": False,
                },
            ]),
        ]

        for archived, expected_results in cases:
            with self.subTest(archived=archived), self.assertNumQueries(4):
                response = self.client.get(reverse("api:models"), {"archived": archived})
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertEqual(response.json()["results"], expected_results)

    def test_list_models_filter_compatible_worker_doesnt_exist(self):
        self.client.force_login(self.user2)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:models"), {"compatible_worker": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        models = response.json()["results"]

        self.assertListEqual(models, [])

    def test_list_models_filter_compatible_worker_no_models(self):
        self.client.force_login(self.user2)
        dla_worker = Worker.objects.get(slug="dla")
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:models"), {"compatible_worker": str(dla_worker.id)})
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        models = response.json()["results"]

        self.assertListEqual(models, [])

    def test_list_models_filter_compatible_worker_invalid_uuid(self):
        self.client.force_login(self.user2)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:models"), {"compatible_worker": "uwu"})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"compatible_worker": ["uwu is not a valid UUID."]})

    # DownloadModel

    def test_download_model_version_wrong_model_version_id(self):
        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:model-version-download", kwargs={"pk": uuid4()}), {"token": "hehe_perhaps"})
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No ModelVersion matches the given query."})

    def test_download_model_version_non_available(self):
        """
        It is not possible to download a model version that has not been validated
        """
        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:model-version-download", kwargs={"pk": str(self.model_version1.id)}), {"token": "test"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "This model version is not available yet."})

    def test_download_model_version_wrong_token(self):
        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:model-version-download", kwargs={"pk": str(self.model_version5.id)}), {"token": "hehe_perhaps"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Wrong token."})

    def test_download_model_version(self):
        token = self.model_version5.build_authentication_token_hash()
        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:model-version-download", kwargs={"pk": str(self.model_version5.id)}), {"token": token})
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.headers["Location"], self.model_version5.s3_url)

    # RetrieveModel

    def test_retrieve_model_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:model-retrieve", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_model_requires_verified(self):
        self.user3.verified_email = False
        self.user3.save()
        self.client.force_login(self.user3)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:model-retrieve", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Model.objects.none())
    def test_retrieve_model_requires_guest(self, filter_rights_mock):
        self.client.force_login(self.user3)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user3, Model, Role.Guest.value))

    def test_retrieve_model(self):
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:model-retrieve", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.model2.id),
            "created": self.model2.created.isoformat().replace("+00:00", "Z"),
            "updated": self.model2.updated.isoformat().replace("+00:00", "Z"),
            "name": "Second Model",
            "description": "",
            "rights": ["read", "write", "admin"],
            "archived": False,
        })

    # UpdateModel

    def test_update_model_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(reverse("api:model-retrieve", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_update_model_requires_verified(self):
        self.user3.verified_email = False
        self.user3.save()
        self.client.force_login(self.user3)
        with self.assertNumQueries(2):
            response = self.client.put(reverse("api:model-retrieve", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_update_model_requires_contributor(self, has_access_mock):
        self.assertFalse(self.model1.public)
        self.model1.memberships.create(user=self.user3, level=Role.Guest.value)
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                {"name": "new name", "description": "test"},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.model1.refresh_from_db()
        self.assertEqual((self.model1.name, self.model1.description), ("First Model", "first"))
        self.assertDictEqual(response.json(), {"detail": "You do not have a contributor access to this model."})
        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user3, self.model1, Role.Contributor.value, skip_public=False))

    def test_update_model_unique_name(self):
        self.client.force_login(self.user1)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                {"name": self.model2.name, "description": "", "public": True},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["A model with this name already exists"]})

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_update_model_archived_requires_archivable(self, get_max_level_mock):
        self.client.force_login(self.user2)
        self.assertFalse(self.model1.is_archivable(self.user2))

        cases = [
            (timezone.now(), False),
            (None, True),
        ]
        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                get_max_level_mock.reset_mock()
                self.model1.archived = current_value
                self.model1.save()

                with self.assertNumQueries(4):
                    response = self.client.put(
                        reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                        {"name": "new name", "description": "test", "archived": new_value},
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {"archived": ["You are not allowed to archive or unarchive this model."]})
                self.assertEqual(get_max_level_mock.call_count, 1)
                self.assertEqual(get_max_level_mock.call_args, call(self.user2, self.model1))

    def test_update_model_archived(self):
        self.assertTrue(self.model1.is_archivable(self.user1))
        self.client.force_login(self.user1)

        cases = [
            (timezone.now(), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                self.model1.archived = current_value
                self.model1.save()

                with self.assertNumQueries(5):
                    response = self.client.put(
                        reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                        {"name": "new name", "description": "test", "archived": new_value},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.model1.refresh_from_db()
                self.assertDictEqual(
                    response.json(),
                    {
                        "id": str(self.model1.id),
                        "description": "test",
                        "name": "new name",
                        "rights": ["read", "write", "admin"],
                        "created": self.model1.created.isoformat().replace("+00:00", "Z"),
                        "updated": self.model1.updated.isoformat().replace("+00:00", "Z"),
                        "archived": new_value,
                    }
                )

    def test_update_model(self):
        self.assertFalse(self.model1.public)
        self.client.force_login(self.user1)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                {"name": "new name", "description": "", "public": True},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.model1.refresh_from_db()
        self.assertEqual((self.model1.name, self.model1.description), ("new name", ""))
        # Public field cannot be edited
        self.assertFalse(self.model1.public)
        self.assertDictEqual(
            response.json(),
            {
                "id": str(self.model1.id),
                "description": "",
                "name": "new name",
                "rights": ["read", "write", "admin"],
                "created": self.model1.created.isoformat().replace("+00:00", "Z"),
                "updated": self.model1.updated.isoformat().replace("+00:00", "Z"),
                "archived": False,
            }
        )

    # PartialUpdateModel

    def test_partial_update_model_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(reverse("api:model-retrieve", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_partial_update_model_requires_verified(self):
        self.user3.verified_email = False
        self.user3.save()
        self.client.force_login(self.user3)

        with self.assertNumQueries(2):
            response = self.client.patch(reverse("api:model-retrieve", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_partial_update_model_requires_contributor(self, has_access_mock):
        self.assertFalse(self.model1.public)
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                {"description": "test"},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.model1.refresh_from_db()
        self.assertEqual((self.model1.name, self.model1.description), ("First Model", "first"))
        self.assertDictEqual(response.json(), {"detail": "You do not have a contributor access to this model."})

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user3, self.model1, Role.Contributor.value, skip_public=False))

    def test_partial_update_model(self):
        self.client.force_login(self.user1)
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                {"description": "new desc", "public": True},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.model1.refresh_from_db()
        self.assertEqual((self.model1.name, self.model1.description), ("First Model", "new desc"))
        # Public field cannot be edited
        self.assertFalse(self.model1.public)
        self.assertDictEqual(
            response.json(),
            {
                "id": str(self.model1.id),
                "description": "new desc",
                "name": "First Model",
                "rights": ["read", "write", "admin"],
                "created": self.model1.created.isoformat().replace("+00:00", "Z"),
                "updated": self.model1.updated.isoformat().replace("+00:00", "Z"),
                "archived": False,
            }
        )

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_partial_update_model_archived_requires_archivable(self, get_max_level_mock):
        self.assertFalse(self.model1.is_archivable(self.user2))
        self.client.force_login(self.user2)

        cases = [
            (timezone.now(), False),
            (None, True),
        ]
        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                get_max_level_mock.reset_mock()
                self.model1.archived = current_value
                self.model1.save()

                with self.assertNumQueries(3):
                    response = self.client.patch(
                        reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                        {"archived": new_value},
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {"archived": ["You are not allowed to archive or unarchive this model."]})
                self.assertEqual(get_max_level_mock.call_count, 1)
                self.assertEqual(get_max_level_mock.call_args, call(self.user2, self.model1))

    def test_partial_update_model_archived(self):
        self.assertTrue(self.model1.is_archivable(self.user1))
        self.client.force_login(self.user1)

        cases = [
            (timezone.now(), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                self.model1.archived = current_value
                self.model1.save()

                with self.assertNumQueries(4):
                    response = self.client.patch(
                        reverse("api:model-retrieve", kwargs={"pk": str(self.model1.id)}),
                        {"archived": new_value},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.model1.refresh_from_db()
                self.assertDictEqual(
                    response.json(),
                    {
                        "id": str(self.model1.id),
                        "description": "first",
                        "name": "First Model",
                        "rights": ["read", "write", "admin"],
                        "created": self.model1.created.isoformat().replace("+00:00", "Z"),
                        "updated": self.model1.updated.isoformat().replace("+00:00", "Z"),
                        "archived": new_value,
                    }
                )

    # CreateModelVersion

    def test_create_model_version_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)

        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:model-versions", kwargs={"pk": str(self.model2.id)}),
                {
                    "tag": "TAG",
                    "description": "description",
                    "configuration": {"hello": "this is me"},
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.training.api.get_max_level", return_value=Role.Guest.value)
    def test_create_model_version_requires_contributor(self, get_max_level_mock):
        """
        Can't create model version as guest
        """
        self.client.force_login(self.user1)

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:model-versions", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the model to create a new version."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user1, self.model2))

    def test_create_model_version_archived(self):
        self.model1.archived = timezone.now()
        self.model1.save()
        self.client.force_login(self.user1)

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"model": ["This model is archived."]})

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    def test_create_model_version_empty_fields(self, s3_presigned_url_mock):
        """
        Creates a new model version without setting a tag nor any field.
        """
        s3_presigned_url_mock.return_value = "http://s3/upload_put_url"
        self.client.force_login(self.user1)

        fake_now = timezone.now()
        # To mock the creation date
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = fake_now
            with self.assertNumQueries(4):
                response = self.client.post(
                    reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                    {},
                )
                self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn("id", data)
        df = ModelVersion.objects.get(id=data["id"])

        self.assertEqual(df.tag, None)
        self.assertDictEqual(
            data,
            {
                "id": str(df.id),
                "model_id": str(self.model1.id),
                "parent": None,
                "description": "",
                "state": ModelVersionState.Created.value,
                "configuration": {},
                "tag": None,
                "size": None,
                "hash": None,
                "created": fake_now.isoformat().replace("+00:00", "Z"),
                "s3_url": None,
                "s3_put_url": s3_presigned_url_mock.return_value,
            }
        )

    def test_create_model_version_blank_tag_model(self):
        """
        Raise 400 when creating a model version with a blank tag
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                {"tag": ""},
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"tag": ["This field may not be blank."]})

    def test_create_model_version_existing_tag_model(self):
        """
        Raise 400 when creating a model version with an existing tag for the same model
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                {"tag": "tagged"},
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A version for this model with this tag already exists."]})

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    def test_create_model_version(self, s3_presigned_url_mock):
        """
        Creates a new model version with a tag
        """
        self.client.force_login(self.user1)
        s3_presigned_url_mock.return_value = "http://s3/upload_put_url"
        fake_now = timezone.now()
        # To mock the creation date
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = fake_now
            with self.assertNumQueries(6):
                response = self.client.post(
                    reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                    {
                        "tag": "TAG",
                        "size": 12,
                        "hash": "d" * 32,
                        "archive_hash": "e" * 32,
                        "description": "description",
                        "configuration": {"hello": "this is me"},
                    },
                )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn("id", data)
        df = ModelVersion.objects.get(id=data["id"])

        self.assertDictEqual(
            data,
            {
                "id": str(df.id),
                "model_id": str(self.model1.id),
                "parent": None,
                "description": "description",
                "state": ModelVersionState.Created.value,
                "configuration": {"hello": "this is me"},
                "tag": "TAG",
                "size": 12,
                "hash": "d" * 32,
                "created": fake_now.isoformat().replace("+00:00", "Z"),
                "s3_url": None,
                "s3_put_url": s3_presigned_url_mock.return_value
            }
        )

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    def test_create_huge_model_version(self, s3_presigned_url_mock):
        """
        Creates a new model version that is very big
        """
        self.client.force_login(self.user1)
        s3_presigned_url_mock.return_value = "http://s3/upload_put_url"
        fake_now = timezone.now()
        # To mock the creation date
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = fake_now
            with self.assertNumQueries(6):
                response = self.client.post(
                    reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                    {
                        "tag": "TAG",
                        "size": 9223372036854775806,
                        "hash": "d" * 32,
                        "archive_hash": "e" * 32,
                        "description": "description",
                        "configuration": {"hello": "this is me"},
                    },
                )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn("id", data)
        df = ModelVersion.objects.get(id=data["id"])

        self.assertDictEqual(
            data,
            {
                "id": str(df.id),
                "model_id": str(self.model1.id),
                "parent": None,
                "description": "description",
                "state": ModelVersionState.Created.value,
                "configuration": {"hello": "this is me"},
                "tag": "TAG",
                "size": 9223372036854775806,
                "hash": "d" * 32,
                "created": fake_now.isoformat().replace("+00:00", "Z"),
                "s3_url": None,
                "s3_put_url": s3_presigned_url_mock.return_value
            }
        )

    def test_create_model_version_unique_tag_model(self):
        """
        Raises 400 when creating a model version that already exists, same model_id and tag
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                {"tag": self.model_version2.tag},
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A version for this model with this tag already exists."]})

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    def test_create_model_version_any_parent_model_version(self, s3_presigned_url_mock):
        """
        Any readable model version can be set as parent of a model version, not just versions of the same model
        """
        self.client.force_login(self.user1)
        s3_presigned_url_mock.return_value = "http://s3/upload_put_url"
        fake_now = timezone.now()
        # To mock the creation date
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = fake_now
            with self.assertNumQueries(6):
                response = self.client.post(
                    reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                    {
                        "tag": "TAG",
                        "description": "description",
                        "configuration": {"hello": "this is me"},
                        # self.model_version3 belongs to self.model2
                        "parent": str(self.model_version3.id)
                    },
                )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn("id", data)
        df = ModelVersion.objects.get(id=data["id"])

        self.assertDictEqual(
            data,
            {
                "id": str(df.id),
                "model_id": str(self.model1.id),
                "parent": str(self.model_version3.id),
                "description": "description",
                "state": ModelVersionState.Created.value,
                "configuration": {"hello": "this is me"},
                "tag": "TAG",
                "size": None,
                "hash": None,
                "created": fake_now.isoformat().replace("+00:00", "Z"),
                "s3_url": None,
                "s3_put_url": s3_presigned_url_mock.return_value
            }
        )

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_create_model_version_readable_parent_version(self, filter_rights_mock):
        """
        Only model versions the user has read access to can be set as parents of a model version
        """
        filter_rights_mock.return_value = Model.objects.filter(id=self.model1.id)
        self.client.force_login(self.user1)
        fake_now = timezone.now()
        # To mock the creation date
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = fake_now
            with self.assertNumQueries(4):
                response = self.client.post(
                    reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                    {
                        "tag": "TAG",
                        "description": "description",
                        "configuration": {"hello": "this is me"},
                        # self.model_version3 belongs to self.model2
                        "parent": str(self.model_version3.id)
                    },
                )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user1, Model, Role.Guest.value))
        self.assertDictEqual(response.json(), {
            "parent": [f'Invalid pk "{str(self.model_version3.id)}" - object does not exist.']
        })

    def test_create_model_version_unique_hash(self):
        self.client.force_login(self.user1)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}),
                {"hash": self.model_version5.hash},
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "hash": ["A model version with this hash already exists."],
            "id": [str(self.model_version5.id)]
        })

    # ListModelVersions

    def test_list_model_versions_requires_logged_in(self):
        """To list a model's versions, you need to be logged in.
        """
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:model-versions", kwargs={"pk": str(self.model1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    @patch("arkindex.training.api.get_max_level", return_value=None)
    def test_list_model_versions_requires_guest_access(self, get_max_level_mock):
        self.client.force_login(self.user1)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:model-versions", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You need a Guest access to list versions of a model."})

    @patch("arkindex.training.api.get_max_level", return_value=Role.Guest.value)
    def test_list_model_versions_low_access(self, get_max_level_mock):
        """With only guest access rights on a model, you only see the available versions with a set tag
        """
        self.client.force_login(self.user3)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:model-versions", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            [version["id"] for version in response.json()["results"]],
            [str(self.model_version3.id)],
        )

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user3, self.model2))

    @patch("arkindex.training.api.get_max_level", return_value=Role.Contributor.value)
    def test_list_model_versions_full_access(self, get_max_level_mock):
        """With contributor access rights, you see all versions.
        """
        self.client.force_login(self.user2)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:model-versions", kwargs={"pk": str(self.model2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            [version["id"] for version in response.json()["results"]],
            [str(self.model_version4.id), str(self.model_version3.id)],
        )

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user2, self.model2))

    # DetroyModelVersion

    @patch("arkindex.training.api.get_max_level", return_value=Role.Contributor.value)
    def test_destroy_model_versions_requires_admin(self, get_max_level_mock):
        self.client.force_login(self.user2)

        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You need an Admin access to the model to destroy this version."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user2, self.model1))

    def test_destroy_model_versions_archived(self):
        self.model1.archived = timezone.now()
        self.model1.save()
        self.client.force_login(self.user1)

        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertListEqual(response.json(), ["This model is archived."])

    def test_destroy_model_versions(self):
        """To destroy a model version, you need admin rights on the model.
        This also deletes every worker run that used this model version
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(8):
            response = self.client.delete(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    # RetrieveModelversion

    @patch("arkindex.training.api.get_max_level", return_value=None)
    def test_retrieve_model_versions_require_guest(self, get_max_level_mock):
        """
        To retrieve a model version with a set tag and state==Available, you need guest rights on the model.
        """
        self.client.force_login(self.user1)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You need a Guest access to the model to retrieve this version."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user1, self.model2))

    @patch("arkindex.training.api.get_max_level", return_value=Role.Guest.value)
    def test_retrieve_model_versions_tag_available(self, get_max_level_mock):
        """
        Retrieve a model version with a set tag and state==Available with guest rights on the model.
        No s3 URL is set with a guest access.
        """
        self.assertIsNotNone(self.model_version3.tag)
        self.assertEqual(self.model_version3.state, ModelVersionState.Available)
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.model_version3.id),
            "model_id": str(self.model2.id),
            "parent": None,
            "description": "",
            "tag": "tagged",
            "hash": "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbba",
            "state": "available",
            "size": 8,
            "configuration": {},
            "created": self.model_version3.created.isoformat().replace("+00:00", "Z"),
            "s3_url": None,
            "s3_put_url": None,
        })
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user3, self.model2))

    @patch("arkindex.training.api.get_max_level", return_value=Role.Guest.value)
    def test_retrieve_model_versions_no_tag_requires_contributor(self, get_max_level_mock):
        """
        To retrieve a model version with no set tag or state!=Available, you need contributor rights on the model.
        """
        self.assertEqual(self.model_version4.tag, None)
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version4.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the model to retrieve this version."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user3, self.model2))

    @patch("arkindex.training.api.get_max_level", return_value=Role.Guest.value)
    def test_retrieve_model_versions_no_validated_requires_contributor(self, get_max_level_mock):
        """
        To retrieve a model version with no set tag or state!=Available, you need contributor rights on the model.
        """
        self.model_version4.tag = "A tag"
        self.model_version4.state = ModelVersionState.Error
        self.model_version4.save()
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version4.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the model to retrieve this version."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user3, self.model2))

    @patch("arkindex.training.api.get_max_level", return_value=Role.Contributor.value)
    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url", return_value="http://s3/get_url")
    def test_retrieve_model_versions(self, s3_presigned_url, get_max_level_mock):
        """
        Retrieve a model version with no set tag or state!=Available with contributor rights on the model.
        """
        self.client.force_login(self.user2)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version4.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.model_version4.id),
            "model_id": str(self.model2.id),
            "parent": None,
            "description": "some description",
            "tag": None,
            "hash": "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbaa",
            "state": ModelVersionState.Available.value,
            "size": 8,
            "configuration": {"n_epochs": 10},
            "created": self.model_version4.created.isoformat().replace("+00:00", "Z"),
            "s3_url": "http://s3/get_url",
            "s3_put_url": None,
        })

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user2, self.model2))

    # PartialUpdateModelVersion

    @patch("arkindex.training.api.get_max_level", return_value=Role.Guest.value)
    def test_partial_update_model_version_requires_contributor(self, get_max_level_mock):
        """
        Can't partial update a model version as guest
        """
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.patch(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}), {"state": "available"})
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(
            response.json(),
            {"detail": "You need a Contributor access to the model to update this version."}
        )

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user3, self.model2))

    @patch("arkindex.project.aws.s3.Object")
    @patch("arkindex.project.aws.S3FileMixin.exists")
    def test_partial_update_model_version_archived(self, exists, s3_object):
        s3_object().content_length = self.model_version3.size
        s3_object().e_tag = self.model_version3.archive_hash
        exists.return_value = True
        self.model1.archived = timezone.now()
        self.model1.save()
        self.client.force_login(self.user1)

        with self.assertNumQueries(3):
            response = self.client.patch(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}), {"state": "available"})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"model": ["This model is archived."]})

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    @patch("arkindex.project.aws.S3FileMixin.exists")
    @patch("arkindex.project.aws.s3.Object")
    def test_partial_update_created_model_version(self, s3_object, exists, s3_presigned_url):
        s3_presigned_url.return_value = "http://s3/upload_url"
        exists.return_value = True
        s3_object().content_length = self.model_version1.size
        s3_object().e_tag = self.model_version1.archive_hash
        self.client.force_login(self.user1)
        request = {
            "description" : "A very long description",
            "tag" : "prod",
            "configuration": {"parameter1": "value1"},
            "parent": str(self.model_version5.id),
            "state" : ModelVersionState.Created.value,
            "size": 1337,
            "hash": "c" * 32,
            "archive_hash": "c" * 32,
        }

        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}),
                request,
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.model_version1.id),
            "model_id": str(self.model_version1.model_id),
            "created": self.model_version1.created.isoformat().replace("+00:00", "Z"),
            "s3_url": None,
            "s3_put_url": "http://s3/upload_url",
            "state": ModelVersionState.Created.value,
            # Modified attributes
            "tag": "prod",
            "description": request["description"],
            "configuration": request["configuration"],
            "parent": str(self.model_version5.id),
            "hash": "c" * 32,
            "size": 1337,
        })
        self.model_version1.refresh_from_db()
        self.assertEqual(self.model_version1.archive_hash, "c" * 32)

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    @patch("arkindex.project.aws.S3FileMixin.exists")
    @patch("arkindex.project.aws.s3.Object")
    def test_partial_update_available_model_version(self, s3_object, exists, s3_presigned_url):
        s3_presigned_url.return_value = "http://s3/get_url"
        exists.return_value = True
        s3_object().content_length = 1337
        s3_object().e_tag = self.model_version4.archive_hash
        self.client.force_login(self.user2)
        request = {
            "description" : "A very long description",
            "tag" : "prod",
            "configuration": {"parameter1": "value1"},
            "parent": str(self.model_version4.id),
            "state" : ModelVersionState.Available.value,
            "size": 1337,
            "hash": "c" * 32,
            "archive_hash": self.model_version4.archive_hash,
        }

        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}),
                request,
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())

        self.assertDictEqual(response.json(), {
            "id": str(self.model_version3.id),
            "model_id": str(self.model_version3.model_id),
            "created": self.model_version3.created.isoformat().replace("+00:00", "Z"),
            "s3_url": "http://s3/get_url",
            "s3_put_url": None,
            "state": ModelVersionState.Available.value,
            # Modified attributes
            "tag": "prod",
            "description": request["description"],
            "configuration": request["configuration"],
            "parent": str(self.model_version4.id),
            "hash": "c" * 32,
            "size": 1337,
        })
        self.model_version3.refresh_from_db()
        self.assertEqual(self.model_version3.archive_hash, self.model_version4.archive_hash)

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    def test_partial_update_remove_parent(self, s3_presigned_url):
        """
        The parent of a model version can be removed
        """
        s3_presigned_url.return_value = "http://s3/get_url"
        self.client.force_login(self.user2)

        self.model_version3.parent = self.model_version2
        self.model_version3.save()

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}),
                {"parent": None},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.model_version3.id),
            "model_id": str(self.model_version3.model_id),
            "created": self.model_version3.created.isoformat().replace("+00:00", "Z"),
            "s3_url": "http://s3/get_url",
            "s3_put_url": None,
            "tag": "tagged",
            "description": "",
            "configuration": {},
            "parent": None,
            "state": ModelVersionState.Available.value,
            "hash": "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbba",
            "size": 8,
        })

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    def test_partial_update_any_parent_model_version(self, s3_presigned_url):
        """
        A model version can have any model version as a parent, not just a version from the same model
        """
        s3_presigned_url.return_value = "http://s3/get_url"
        self.client.force_login(self.user2)

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}),
                # self.model_version3 is a version of self.model2, while self.model_version1 belongs to self.model1
                {"parent": str(self.model_version1.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.model_version3.id),
            "model_id": str(self.model2.id),
            "created": self.model_version3.created.isoformat().replace("+00:00", "Z"),
            "s3_url": "http://s3/get_url",
            "s3_put_url": None,
            "tag": "tagged",
            "description": "",
            "configuration": {},
            "parent": str(self.model_version1.id),
            "state": ModelVersionState.Available.value,
            "hash": "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbba",
            "size": 8,
        })

        self.model_version3.refresh_from_db()
        self.assertEqual(self.model_version3.parent_id, self.model_version1.id)

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_partial_update_readable_parent_model_version(self, filter_rights_mock):
        """
        Only model versions the user has read access to can be set as parents of a model version
        """
        filter_rights_mock.return_value = Model.objects.filter(id__in=[self.model2.id, self.model3.id])
        self.client.force_login(self.user2)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}),
                # self.model_version3 is a version of self.model2, while self.model_version1 belongs to self.model1
                {"parent": str(self.model_version1.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user2, Model, Role.Guest.value))
        self.assertDictEqual(response.json(), {
            "parent": [f'Invalid pk "{str(self.model_version1.id)}" - object does not exist.']
        })

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    @patch("arkindex.project.aws.S3FileMixin.exists")
    @patch("arkindex.project.aws.s3.Object")
    def test_partial_update_model_version_unique_hash(self, s3_object, exists, s3_presigned_url):
        s3_presigned_url.return_value = "http://s3/get_url"
        exists.return_value = True
        s3_object().content_length = self.model_version1.size
        s3_object().e_tag = self.model_version1.archive_hash
        self.client.force_login(self.user2)
        current_hash = self.model_version3.hash

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}),
                {"hash": self.model_version4.hash},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "hash": ["A model version with this hash already exists."],
            "id": [str(self.model_version4.id)]
        })
        self.model_version3.refresh_from_db()
        self.assertEqual(self.model_version3.hash, current_hash)

    @patch("arkindex.project.aws.S3FileMixin.exists")
    def test_partial_update_model_version_archive_not_present(self, exists):
        """
        When the archive is not found while updating a model version's hash,
        archive hash, and size, a validation error is raised
        """
        self.client.force_login(self.user1)
        exists.return_value = False
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}),
                {"archive_hash": "d" * 32, "hash": "e" * 32, "size": 32, "state": "available"},
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "size": ["Archive has not been uploaded"],
            "archive_hash": ["No file content, assert file has been correctly uploaded"]
        })
        self.model_version1.refresh_from_db()
        self.assertEqual(self.model_version1.hash, None)
        self.assertEqual(self.model_version1.state, ModelVersionState.Error)

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    @patch("arkindex.project.aws.S3FileMixin.exists")
    @patch("arkindex.project.aws.s3.Object")
    def test_partial_update_model_version_wrong_size(self, s3_object, exists, s3_presigned_url):
        s3_presigned_url.return_value = "http://s3/get_url"
        exists.return_value = True
        s3_object().content_length = 31
        s3_object().e_tag = "d" * 32
        self.client.force_login(self.user1)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}),
                {"archive_hash": "d" * 32, "hash": "e" * 32, "size": 32, "state": "available"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"size": ["Uploaded file size is 31 bytes, expected 32 bytes"]})
        self.model_version1.refresh_from_db()
        self.assertEqual(self.model_version1.hash, None)
        self.assertEqual(self.model_version1.state, ModelVersionState.Error)

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    @patch("arkindex.project.aws.S3FileMixin.exists")
    @patch("arkindex.project.aws.s3.Object")
    def test_partial_update_model_version_wrong_hash(self, s3_object, exists, s3_presigned_url):
        s3_presigned_url.return_value = "http://s3/get_url"
        exists.return_value = True
        s3_object().content_length = 32
        s3_object().e_tag = "a" * 32
        current_hash = self.model_version3.hash

        self.client.force_login(self.user1)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}),
                {"archive_hash": "d" * 32, "hash": "e" * 32, "size": 32},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"archive_hash": ["MD5 hashes do not match"]})
        self.model_version3.refresh_from_db()
        self.assertEqual(self.model_version3.hash, current_hash)
        self.assertEqual(self.model_version3.state, ModelVersionState.Error)

    # UpdateModelVersion

    @patch("arkindex.training.api.get_max_level", return_value=None)
    def test_update_model_version_requires_contributor(self, get_max_level_mock):
        """
        Can't update a model version with guest access rights to the model
        """
        self.client.force_login(self.user3)

        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}),
                {
                    "hash": "94274e84f3de91d1645b1e082b5f3466",
                    "archive_hash": "0958a74b060a89fc38318a9a96aef32a",
                    "size": 8,
                    "state": ModelVersionState.Available.value,
                    "description": "other description",
                    "configuration": {"hi": "Who am I?"},
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the model to update this version."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user3, self.model2))

    @patch("arkindex.project.aws.s3.Object")
    @patch("arkindex.project.aws.S3FileMixin.exists")
    def test_update_model_version_archived(self, exists, s3_object):
        s3_object().content_length = self.model_version3.size
        s3_object().e_tag = self.model_version3.archive_hash
        exists.return_value = True
        self.model1.archived = timezone.now()
        self.model1.save()
        self.client.force_login(self.user1)

        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}),
                {
                    "hash": "94274e84f3de91d1645b1e082b5f3466",
                    "archive_hash": "0958a74b060a89fc38318a9a96aef32a",
                    "size": 8,
                    "state": ModelVersionState.Available.value,
                    "description": "other description",
                    "configuration": {"hi": "Who am I ?"},
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"model": ["This model is archived."]})
